; #########################################################################
;
;                             ZSub
;
;##########################################################################

  ; -----------------------------------------------------------------------
  ; This procedure was written by Raymond Filiatreault, February 2005
  ;
  ;              Src1 - Src2 -> Dest  = (a1-a2),(b1-b2)
  ;
  ; This ZSub function subtracts the complex numbers from two sources (Src1
  ; and Src2) with the FPU and returns the result as a complex number at the
  ; specified destination (the FPU itself or a memory location), unless an
  ; invalid operation is detected or the definition of the parameters (with
  ; uID) is invalid.
  ;
  ; The source can be either in cartesian or polar coordinates. The result
  ; will be returned in the same coordinates.
  ;
  ; Either of the two sources can be a complex number already on the FPU
  ; itself or from memory. If on the FPU, the REAL coefficient must be in
  ; the ST(0) register and the IMAGINARY coefficient in the ST(1) register.
  ; Memory variables must have the two coefficients at adjacent memory
  ; addresses, the REAL coefficient being at the lower address. All memory
  ; sources and destinations must have the same declared size as specified
  ; in the uID parameter.
  ;
  ; None of the sources are checked for validity. This is the programmer's
  ; responsibility. The programmer is also responsible to insure that all
  ; the FPU data registers necessary for performing the function are FREE.
  ;   If either Src1 or Src2 is on the FPU:  ST(6) and ST(7) must be free
  ;   If both sources are on the FPU: no other register needs to be free
  ;   Otherwise: ST(4), ST(5), ST(6) and ST(7) must be free
  ; One more register must be free if sources are in polar coordinates
  ;
  ; Only EAX is used to return error (0) or success (!0). All other CPU
  ; registers are preserved.
  ;
  ; If a valid operand size is present in the uID, any source data on the
  ; FPU will be modified; otherwise, if an invalid operand size is present
  ; in the uID, all FPU registers are preserved and the location of the
  ; destination is not altered.
  ;
  ; If an invalid operation is detected, the modified data would be
  ; considered useless and the top two registers would be popped; also, if
  ; a memory location is specified as the destination, it would not be
  ; modified.
  ;
  ; -----------------------------------------------------------------------

    .386
    .model flat, stdcall  ; 32 bit memory model
    option casemap :none  ; case sensitive

    include ZLIB.inc

    .code

;###########################################################################

ZSub proc public USES ecx lpSrc1:DWORD, lpSrc2:DWORD, lpDest:DWORD, uID:DWORD
        
;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

      mov   ecx,uID
      and   ecx,Z_SINGLE or Z_DBL   ;check uID for operand size
      .if   ZERO?
            mov   ecx,10
            jmp   @F
      .elseif ecx == 4
            jmp   @F
      .elseif ecx == 8
            jmp   @F
      .endif
      xor   eax,eax                 ;error code
      ret                           ;no valid operand size

   @@:
      test  uID,Z_POLAR             ;polar coordinates?
      .if   !ZERO?
            test  uID,SRC1_FPU or SRC2_FPU or SRC12_FPU ;any Src on FPU?
            .if   !ZERO?

                  ;------------------------------------------------------
                  ; Polar coordinates cannot be subtracted directly
                  ; They must first be converted to cartesian coordinates
                  ;------------------------------------------------------
                  
                  invoke ZPtoC,0,0,SRC1_FPU or DEST_FPU  ;->cartesian
            .endif
            test  uID,SRC12_FPU                 ;check if both Src on FPU
            .if   !ZERO?
                                    ;a2   b2   r1   t1
                  fxch  st(2)       ;r1   b2   a2   t1
                  fxch              ;b2   r1   a2   t1
                  fxch  st(3)       ;t1   r1   a2   b2
                  fxch              ;r1   t1   a2   b2
                  invoke ZPtoC,0,0,SRC1_FPU or DEST_FPU   ;->cartesian
                  fxch  st(2)       ;a2   b1   a1   b2
                  fxch              ;b1   a2   a1   b2
                  fxch  st(3)       ;b2   a2   a1   b1
                  fxch              ;a2   b2   a1   b1
            .endif
      .endif

      test  uID,SRC12_FPU
      jnz   subthem

      test  uID,SRC1_FPU            ;check if Src1 in on FPU
      .if   ZERO?                   ;Src1 not on FPU
            mov   eax,lpSrc1
            .if   ecx == 4
                  fld   dword ptr [eax+ecx]     ;imag1
                  fld   dword ptr [eax]         ;real1   imag1
            .elseif ecx == 8
                  fld   qword ptr [eax+ecx]
                  fld   qword ptr [eax]
            .elseif ecx == 10
                  fld   tbyte ptr [eax+ecx]
                  fld   tbyte ptr [eax]
            .endif

            test  uID,Z_POLAR             ;polar coordinates?
            .if   !ZERO?
                  invoke ZPtoC,0,0,SRC1_FPU or DEST_FPU  ;->cartesian
            .endif

      .endif

      test  uID,SRC2_FPU            ;check if Src2 in on FPU
      .if   ZERO?                   ;Src2 not on FPU
            mov   eax,lpSrc2
            .if   ecx == 4
                  fld   dword ptr [eax+ecx]     ;imag2   real1   imag1
                  fld   dword ptr [eax]         ;real2   imag2   real1   imag1
            .elseif ecx == 8
                  fld   qword ptr [eax+ecx]
                  fld   qword ptr [eax]
            .else
                  fld   tbyte ptr [eax+ecx]
                  fld   tbyte ptr [eax]
            .endif

            test  uID,Z_POLAR             ;polar coordinates?
            .if   !ZERO?
                  invoke ZPtoC,0,0,SRC1_FPU or DEST_FPU  ;->cartesian
            .endif
      .else                         ;Src2 was on FPU before loading Src1
            test  uID,SRC1_FPU
            .if   !ZERO?                   ;Src1 also on FPU, make copy
                  fld   st(1)       ;b2   a1    b1
                  fld   st(1)       ;a2   b2   a1    b1
            .else             ;inverse them
                                    ;a1   b1    a2    b2
                  fxch  st(2)       ;a2   b1    a1    b2
                  fxch              ;b1   a2    a1    b2
                  fxch  st(3)       ;b2   a2    a1    b1
                  fxch              ;a2   b2    a1    b1
            .endif
      .endif

subthem:
;==================
; (a1-a2) , (b1-b2)
;==================

      fsubp st(2),st          ;b2    a1-a2   b1
      fsubp st(2),st          ;a1-a2   b1-b2

      fstsw ax                      ;get exception flags
      fwait
      shr   al,1                    ;test for invalid operation
      .if   CARRY?                  ;indicates invalid operation
            fstp  st                ;pop top register
            fstp  st                ;pop second register
            xor   eax,eax           ;error return code
            ret
      .endif

      test  uID,Z_POLAR             ;polar coordinates?
      .if   !ZERO?
            invoke ZCtoP,0,0,SRC1_FPU or DEST_FPU     ;->polar
      .endif

      test  uID,DEST_FPU            ;check destination of result
      .if   ZERO?                   ;to be stored in memory
            mov   eax,lpDest
            .if   ecx == 4
                  fstp  dword ptr [eax]         ;store REAL first
                  fstp  dword ptr [eax+ecx]     ;then store IMAG
            .elseif ecx == 8
                  fstp  qword ptr [eax]
                  fstp  qword ptr [eax+ecx]
            .else
                  fstp  tbyte ptr [eax]
                  fstp  tbyte ptr [eax+ecx]
            .endif
      .endif

      or    al,1                    ;to insure EAX!=0
      ret
    
ZSub endp

; #########################################################################

end
